@@ -1,6 +1,6 @@
 #
-# Author:: Adam Jacob (<adam@opscode.com>)
-# Copyright:: Copyright (c) 2008 Opscode, Inc.
+# Author:: Adam Jacob (<adam@chef.io>)
+# Copyright:: Copyright (c) 2008-2019 Chef Software, Inc.
 # License:: Apache License, Version 2.0
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -17,7 +17,7 @@
 #
 
 require "optparse"
-
+require_relative "cli/formatter"
 module Mixlib
 
   # == Mixlib::CLI
@@ -30,8 +30,9 @@ module Mixlib
   # === DSL
   # When included, Mixlib::CLI also extends the including class with its
   # ClassMethods, which define the DSL. The primary methods of the DSL are
-  # ClassMethods#option, which defines a command line option, and
-  # ClassMethods#banner, which defines the "usage" banner.
+  # ClassMethods#option, which defines a command line option;
+  # ClassMethods#banner, which defines the "usage" banner;
+  # and ClassMethods#deprecated_option, which defines a deprecated command-line option.
   #
   # === Parsing
   # Command line options are parsed by calling the instance method
@@ -41,7 +42,7 @@ module Mixlib
 
     module InheritMethods
       def inherited(receiver)
-        receiver.options = deep_dup(self.options)
+        receiver.options = deep_dup(options)
         receiver.extend(Mixlib::CLI::InheritMethods)
       end
 
@@ -51,8 +52,8 @@ module Mixlib
       # contents will be iterated and cloned as well.
       def deep_dup(object)
         cloned_object = object.respond_to?(:dup) ? object.dup : object
-        if cloned_object.kind_of?(Enumerable)
-          if cloned_object.kind_of?(Hash)
+        if cloned_object.is_a?(Enumerable)
+          if cloned_object.is_a?(Hash)
             new_hash = cloned_object.class.new
             cloned_object.each do |key, value|
               cloned_key = deep_dup(key)
@@ -93,14 +94,89 @@ module Mixlib
       # === Parameters
       # name<Symbol>:: The name of the option to add
       # args<Hash>:: A hash of arguments for the option, specifying how it should be parsed.
-      # === Returns
-      # true:: Always returns true.
+      #   Supported arguments:
+      #     :short   - The short option, just like from optparse. Example: "-l LEVEL"
+      #     :long    - The long option, just like from optparse. Example: "--level LEVEL"
+      #     :description - The description for this item, just like from optparse.
+      #     :default - A default value for this option.  Default values will be populated
+      #     on parse into `config` or `default_default`, depending `use_separate_defaults`
+      #     :boolean - indicates the flag is a boolean. You can use this if the flag takes no arguments
+      #                The config value will be set to 'true' if the flag is provided on the CLI and this
+      #                argument is set to true. The config value will be set to false only
+      #                if it has a default value of false
+      #     :required - When set, the option is required.  If the command is run without this option,
+      #                it will print a message informing the user of the missing requirement, and exit. Default is false.
+      #     :proc     - Proc that will be invoked if the human has specified this option.
+      #                 Two forms are supported:
+      #                 Proc/1 - provided value is passed in.
+      #                 Proc/2 - first argument is provided value. Second is the cli flag option hash.
+      #                 Both versions return the value to be assigned to the option.
+      #     :show_options - this option is designated as one that shows all supported options/help when invoked.
+      #     :exit     - exit your program with the exit code when this option is given. Example: 0
+      #     :in       - array containing a list of valid values. The value provided at run-time for the option is
+      #                 validated against this. If it is not in the list, it will print a message and exit.
+      #     :on :head OR :tail - force this option to display at the beginning or end of the
+      #                          option list, respectively
+      # =
+      # @return <Hash> :: the config hash for the created option
+      # i
       def option(name, args)
         @options ||= {}
-        raise(ArgumentError, "Option name must be a symbol") unless name.kind_of?(Symbol)
+        raise(ArgumentError, "Option name must be a symbol") unless name.is_a?(Symbol)
+
         @options[name.to_sym] = args
       end
 
+      # Declare a deprecated option
+      #
+      # Add a deprecated command line option.
+      #
+      # name<Symbol> :: The name of the deprecated option
+      # replacement<Symbol> :: The name of the option that replaces this option.
+      # long<String> :: The original long flag name, or flag name with argument, eg "--user USER"
+      # short<String>  :: The original short-form flag name, eg "-u USER"
+      # boolean<String> :: true if this is a boolean flag, eg "--[no-]option".
+      # value_mapper<Proc/1> :: a block that accepts the original value from the deprecated option,
+      #                   and converts it to a value suitable for the new option.
+      #                   If not provided, the value provided to the deprecated option will be
+      #                   assigned directly to the converted option.
+      # keep<Boolean> :: Defaults to true, this ensure sthat `options[:deprecated_flag]` is
+      #                  populated when the deprecated flag is used. If set to false,
+      #                  only the value in `replacement` will be set.  Results undefined
+      #                  if no replacement is provided. You can use this to enforce the transition
+      #                  to non-deprecated keys in your code.
+      #
+      # === Returns
+      # <Hash> :: The config hash for the created option.
+      def deprecated_option(name,
+        replacement: nil,
+        long: nil,
+        short: nil,
+        boolean: false,
+        value_mapper: nil,
+        keep: true)
+
+        description = if replacement
+                        replacement_cfg = options[replacement]
+                        display_name = CLI::Formatter.combined_option_display_name(replacement_cfg[:short], replacement_cfg[:long])
+                        "This flag is deprecated. Use #{display_name} instead."
+                      else
+                        "This flag is deprecated and will be removed in a future release."
+                      end
+        value_mapper ||= Proc.new { |v| v }
+
+        option(name,
+          long: long,
+          short: short,
+          boolean: boolean,
+          description: description,
+          on: :tail,
+          deprecated: true,
+          keep: keep,
+          replacement: replacement,
+          value_mapper: value_mapper)
+      end
+
       # Get the hash of current options.
       #
       # === Returns
@@ -118,7 +194,8 @@ module Mixlib
       # === Returns
       # @options<Hash>:: The current options hash.
       def options=(val)
-        raise(ArgumentError, "Options must recieve a hash") unless val.kind_of?(Hash)
+        raise(ArgumentError, "Options must recieve a hash") unless val.is_a?(Hash)
+
         @options = val
       end
 
@@ -181,9 +258,9 @@ module Mixlib
     # === Returns
     # object<Mixlib::Config>:: Returns an instance of whatever you wanted :)
     def initialize(*args)
-      @options = Hash.new
-      @config  = Hash.new
-      @default_config = Hash.new
+      @options = {}
+      @config  = {}
+      @default_config = {}
       @opt_parser = nil
 
       # Set the banner
@@ -209,8 +286,7 @@ module Mixlib
         config_opts[:show_options] ||= false
         config_opts[:exit] ||= nil
         config_opts[:in] ||= nil
-
-        if config_opts.has_key?(:default)
+        if config_opts.key?(:default)
           defaults_container[config_key] = config_opts[:default]
         end
       end
@@ -225,25 +301,31 @@ module Mixlib
     #
     # === Returns
     # argv<Array>:: Returns any un-parsed elements.
-    def parse_options(argv = ARGV)
+    def parse_options(argv = ARGV, show_deprecations: true)
       argv = argv.dup
       opt_parser.parse!(argv)
+      # Do this before our custom validations, so that custom
+      # validations apply to any converted deprecation values;
+      # but after parse! so that everything is populated.
+      handle_deprecated_options(show_deprecations)
 
       # Deal with any required values
-      options.each do |opt_key, opt_value|
-        if opt_value[:required] && !config.has_key?(opt_key)
-          reqarg = opt_value[:short] || opt_value[:long]
+      options.each do |opt_key, opt_config|
+        if opt_config[:required] && !config.key?(opt_key)
+          reqarg = opt_config[:short] || opt_config[:long]
           puts "You must supply #{reqarg}!"
           puts @opt_parser
           exit 2
         end
-        if opt_value[:in]
-          unless opt_value[:in].kind_of?(Array)
+        if opt_config[:in]
+          unless opt_config[:in].is_a?(Array)
             raise(ArgumentError, "Options config key :in must receive an Array")
           end
-          if !opt_value[:in].include?(config[opt_key])
-            reqarg = opt_value[:short] || opt_value[:long]
-            puts "#{reqarg}: #{config[opt_key]} is not included in the list ['#{opt_value[:in].join("', '")}'] "
+
+          if config[opt_key] && !opt_config[:in].include?(config[opt_key])
+            reqarg = Formatter.combined_option_display_name(opt_config[:short], opt_config[:long])
+            puts "#{reqarg}: #{config[opt_key]} is not one of the allowed values: #{Formatter.friendly_opt_list(opt_config[:in])}"
+            # TODO - get rid of this. nobody wants to be spammed with a  ton of information, particularly since we just told them the exact problem and how to fix it.
             puts @opt_parser
             exit 2
           end
@@ -267,7 +349,6 @@ module Mixlib
         # Create new options
         options.sort { |a, b| a[0].to_s <=> b[0].to_s }.each do |opt_key, opt_val|
           opt_args = build_option_arguments(opt_val)
-
           opt_method = case opt_val[:on]
                        when :on
                          :on
@@ -280,7 +361,7 @@ module Mixlib
                        end
 
           parse_block =
-            Proc.new() do |c|
+            Proc.new do |c|
               config[opt_key] = if opt_val[:proc]
                                   if opt_val[:proc].arity == 2
                                     # New hotness to allow for reducer-style procs.
@@ -305,15 +386,59 @@ module Mixlib
       end
     end
 
-    def build_option_arguments(opt_setting)
-      arguments = Array.new
+    # Iterates through options declared as deprecated,
+    # maps values to their replacement options,
+    # and prints deprecation warnings.
+    #
+    # @return NilClass
+    def handle_deprecated_options(show_deprecations)
+      merge_in_values = {}
+      config.each_key do |opt_key|
+        opt_cfg = options[opt_key]
 
-      arguments << opt_setting[:short] if opt_setting.has_key?(:short)
-      arguments << opt_setting[:long] if opt_setting.has_key?(:long)
-      if opt_setting.has_key?(:description)
-        description = opt_setting[:description]
+        # Deprecated entries do not have defaults so no matter what
+        # separate_default_options are set, if we see a 'config'
+        # entry that contains a deprecated indicator, then the option was
+        # explicitly provided by the caller.
+        #
+        # opt_cfg may not exist if an inheriting application
+        # has directly inserted values info config.
+        next unless opt_cfg && opt_cfg[:deprecated]
+
+        replacement_key = opt_cfg[:replacement]
+        if replacement_key
+          # This is the value passed into the deprecated flag. We'll use
+          # the declared value mapper (defaults to return the same value if caller hasn't
+          # provided a mapper).
+          deprecated_val = config[opt_key]
+
+          # We can't modify 'config' since we're iterating it, apply updates
+          # at the end.
+          merge_in_values[replacement_key] = opt_cfg[:value_mapper].call(deprecated_val)
+          config.delete(opt_key) unless opt_cfg[:keep]
+        end
+
+        # Warn about the deprecation.
+        if show_deprecations
+          # Description is also the deprecation message.
+          display_name = CLI::Formatter.combined_option_display_name(opt_cfg[:short], opt_cfg[:long])
+          puts "#{display_name}: #{opt_cfg[:description]}"
+        end
+      end
+      config.merge!(merge_in_values)
+      nil
+    end
+
+    def build_option_arguments(opt_setting)
+      arguments = []
+
+      arguments << opt_setting[:short] if opt_setting[:short]
+      arguments << opt_setting[:long] if opt_setting[:long]
+      if opt_setting.key?(:description)
+        description = opt_setting[:description].dup
         description << " (required)" if opt_setting[:required]
-        description << " (included in ['#{opt_setting[:in].join("', '")}'])" if opt_setting[:in]
+        description << " (valid options: #{Formatter.friendly_opt_list(opt_setting[:in])})" if opt_setting[:in]
+        opt_setting[:description] = description
         arguments << description
       end
 
@@ -324,6 +449,5 @@ module Mixlib
       receiver.extend(Mixlib::CLI::ClassMethods)
       receiver.extend(Mixlib::CLI::InheritMethods)
     end
-
   end
 end
